#include <cmath>
#include <string>
#include <fstream>
#include <streambuf>
#include <algorithm>

#include <libreborn/libreborn.h>
#include <symbols/minecraft.h>

#include <mods/init/init.h>
#include <mods/misc/misc.h>
#include <mods/chat/chat.h>
#include <mods/server/server.h>
#include <mods/feature/feature.h>

static bool compat_mode = true;
static std::string getBlocks(CommandServer *commandserver, Vec3 start, Vec3 end) {
    int startx = start.x, starty = start.y, startz = start.z;
    int endx = end.x, endy = end.y, endz = end.z;

    // Offset cords
    commandserver->pos_translator.from(startx, starty, startz);
    commandserver->pos_translator.from(endx, endy, endz);

    if (endx < startx) std::swap(startx, endx);
    if (endy < starty) std::swap(starty, endy);
    if (endz < startz) std::swap(startz, endz);

    // Get
    std::string ret = "";
    for (int x = startx; x <= endx; x++) {
        for (int y = starty; y <= endy; y++) {
            for (int z = startz; z <= endz; z++) {
                ret += std::to_string(commandserver->minecraft->level->getTile(x, y, z)) + ",";
            }
        }
    }
    if (ret.size() > 1) ret.pop_back();
    return ret + "\n";
}

static void setDir(Entity *entity, float x, float y, float z) {
    if (entity == NULL) return;

    constexpr double _2PI = 2 * M_PI;
    if (x == 0 && z == 0) {
        entity->pitch = y > 0 ? -90 : 90;
        return;
    }

    entity->yaw = fmod((float) atan2(-x, z), (float) _2PI) * (180.0f / M_PI);
    float xz = sqrt(x * x + z * z);
    entity->pitch = (float) atan(-y / xz) * (180.0f / M_PI);
}
static Vec3 getDir(Entity *entity) {
    float y  = -sin(entity->pitch * (M_PI / 180));
    float xz =  cos(entity->pitch * (M_PI / 180));
    float x = -xz * sin(entity->yaw * (M_PI / 180));
    float z =  xz * cos(entity->yaw * (M_PI / 180));
    return Vec3{x, y, z};
}

static std::string empty_fallback(std::string str, int i) {
    if (str == "") return std::to_string(i);
    return str;
}

static std::string getEntityData(CommandServer *commandserver, Entity *entity) {
    float x = entity->x, y = entity->y - entity->height_offset, z = entity->z;
    commandserver->pos_translator.to(x, y, z);
    return std::to_string(entity->id) + "," +
        // type
        std::to_string(entity->getEntityTypeId()) + "," +
        // name
        empty_fallback(misc_get_entity_name_upper(entity, true), entity->id) + "," +
        // x
        std::to_string(x) + "," +
        // y
        std::to_string(y) + "," +
        // z
        std::to_string(z);
}

static float distance_between(Entity *e1, Entity *e2) {
    if (e1 == NULL || e2 == NULL) return 0;
    float dx = e2->x - e1->x;
    float dy = e2->y - e1->y;
    float dz = e2->z - e1->z;
    return sqrt(dx * dx + dy * dy + dz * dz);
}

struct ProjectileHitEvent {
    int x, y, z;
    std::string owner = "";
    int owner_id = 0;
    int targetId = 0;
    bool invalid = false;

    std::string toString(CommandServer *commandserver) {
        float nx = x, ny = y, nz = z;
        commandserver->pos_translator.to(nx, ny, nz);
        return std::to_string((int) nx) + ","
             + std::to_string((int) ny) + ","
             + std::to_string((int) nz) + ","
             + (compat_mode ? "1," : "")
             + (compat_mode ? owner : std::to_string(owner_id)) + "," + std::to_string(targetId);
    }
};

struct ChatEvent {
    std::string message = "";
    int owner_id = 0;
    bool invalid = false;

    std::string toString(__attribute__((unused)) CommandServer *commandserver) {
        return std::to_string(owner_id) + "," + message;
    }
};

static constexpr int EVENT_SIZE = 50;
template <typename T, int SIZE = EVENT_SIZE>
struct CircularQueue {
    int at = 0;
    int start = 0;
    T buf[SIZE];

    void push(T event) {
        buf[at] = event;
        advance_counter(at);
        if (at == start) advance_counter(start);
    }

    T pop() {
        if (start == at) {
            ERR("Over popped projectile hit events!");
        }
        T ret = buf[start];
        advance_counter(start);
        return ret;
    }

    void advance_counter(int &i) {
        i += 1;
        i %= EVENT_SIZE;
    }
};
static CircularQueue<ProjectileHitEvent> hitEvents{};
static CircularQueue<ChatEvent> chatEvents{};

static void clearNewEvents() {
    // Clear chat events
    chatEvents.at = chatEvents.start;
    // Clear projectile events
    if (!compat_mode) hitEvents.at = hitEvents.start;
}

static void clearEvents(CommandServer *commandserver, int id) {
    // Clear block hit events
    if (commandserver->minecraft->isCreativeMode()) {
        EventList_TileEvent &tileEvents = commandserver->minecraft->getCreator()->getTileEvents();
        for (int i = tileEvents.at + 1, j = 0; j < tileEvents.len; i++, j++) {
            if (i == tileEvents.len) i = 0;
            if (tileEvents.events[i].item.entityId == id)
                // This will make the API ignore it
                tileEvents.events[i].tick = -1;
        }
    }
    // Clear chat events
    int i = chatEvents.start;
    while (chatEvents.at != i) {
        if (chatEvents.buf[i].owner_id == id) {
            chatEvents.buf[i].invalid = true;
        }
    }
    // Clear projectile events
    i = hitEvents.start;
    while (hitEvents.at != i) {
        if (hitEvents.buf[i].owner_id == id) {
            hitEvents.buf[i].invalid = true;
        }
    }
}

static std::string getBlockHits(CommandServer *commandserver, ConnectedClient &client, int id) {
    std::string ret = "";
    if (commandserver->minecraft->isCreativeMode()) {
        EventList_TileEvent &tileEvents = commandserver->minecraft->getCreator()->getTileEvents();
        for (int i = tileEvents.at + 1, j = 0; j < tileEvents.len; i++, j++) {
            if (i == tileEvents.len) i = 0;
            if (tileEvents.events[i].item.entityId == id && tileEvents.events[i].tick > client.lastBlockHitPoll) {
                tileEvents.events[i].tick = -1;
                float x = tileEvents.events[i].item.x;
                float y = tileEvents.events[i].item.y;
                float z = tileEvents.events[i].item.z;
                commandserver->pos_translator.to(x, y, z);
                ret +=
                    std::to_string((int) x) + "," +
                    std::to_string((int) y) + "," +
                    std::to_string((int) z) + "," +
                    std::to_string(tileEvents.events[i].item.face) + "," +
                    std::to_string(tileEvents.events[i].item.entityId) + "|";
            }
        }
    }
    if (ret.size() > 1) ret.pop_back();
    return ret + "\n";
}

template<typename T>
static std::string getEventsOfId(CircularQueue<T> events, CommandServer *commandserver, int id) {
    std::string ret = "";
    int i = events.start;
    while (events.at != i) {
        T event = events.buf[i];
        if (event.owner_id == id && !event.invalid) {
            events.buf[i].invalid = true;
            std::string message = event.toString(commandserver);
            if (compat_mode) std::replace(message.begin(), message.end(), '|', '\\');
            else message = misc_base64_encode(message);
            ret += message + "|";
        }
        events.advance_counter(i);
    }
    if (ret.size() > 1) ret.pop_back();
    return ret + "\n";
}

static const std::string fail = "Fail\n";
std::unordered_map<int, int> modern_entity_id_mapping = {
    {93, 10}, /* Chicken */
    {92, 11}, /* Cow */
    {90, 12}, /* Pig */
    {91, 13}, /* Sheep */
    {54, 32}, /* Zombie */
    {50, 33}, /* Creeper */
    {51, 34}, /* Skeleton */
    {52, 35}, /* Spider */
    {57, 36}, /* PigZombie */
    /* ItemEntity isn't available */
    {20, 65}, /* PrimedTnt */
    /* FallingTile isn't available */
    {10, 80}, /* Arrow */
    {11, 81}, /* Snowball */
    {7, 82}, /* Egg */
    {9, 83}, /* Painting */
};
std::string CommandServer_parse_injection(CommandServer_parse_t old, CommandServer *commandserver, ConnectedClient &client, const std::string &command) {
    size_t arg_start = command.find("(");
    if (arg_start == std::string::npos) return fail;
    std::string cmd = command.substr(0, arg_start);

    // rfind so ) in input doesn't break
    size_t cmd_end = command.rfind(")");
    if (cmd_end == std::string::npos) return fail;
    std::string args = command.substr(arg_start + 1, cmd_end - arg_start - 1);

    // And now the big if-else chain
    if (cmd == "world.getBlocks") {
        int x0, y0, z0, x1, y1, z1;
        int ret = sscanf(args.c_str(), "%d,%d,%d,%d,%d,%d", &x0, &y0, &z0, &x1, &y1, &z1);
        if (ret != 6) return fail;

        // Get the blocks
        return getBlocks(commandserver, Vec3{(float) x0, (float) y0, (float) z0}, Vec3{(float) x1, (float) y1, (float) z1});
    } else if (cmd == "world.getPlayerId") {
        for (Player *player : commandserver->minecraft->level->players) {
            if (misc_get_player_username(player) == args)
                return std::to_string(player->id) + "\n";
        }
        WARN("Player [%s] not found.", args.c_str());
        return fail;
    } else if (cmd == "entity.getName") {
        int id;
        int ret = sscanf(args.c_str(), "%d", &id);
        if (ret != 1) return "\n";

        Entity *entity = commandserver->minecraft->level->getEntity(id);
        if (entity == NULL) {
            WARN("Player (or Entity) [%i] not found in entity.getName.", id);
        }
        return empty_fallback(misc_get_entity_name_upper(entity, true), entity->id) + "\n";
    } else if (cmd == "world.getEntities") {
        int type;
        int ret = sscanf(args.c_str(), "%d", &type);
        if (ret != 1) return "\n";

        std::string result = "";
        for (Entity *entity : commandserver->minecraft->level->entities) {
            if (type == -1 || entity->getEntityTypeId() == type) {
                // Get pos
                // Put it together
                result += getEntityData(commandserver, entity) + "|";
            }
        }
        if (result.size() > 1) result.pop_back();
        return result + "\n";
    } else if (cmd == "world.removeEntity") {
        int id;
        int ret = sscanf(args.c_str(), "%d", &id);
        if (ret != 1) return "0\n";

        Entity *entity = commandserver->minecraft->level->getEntity(id);
        if (entity != NULL && !entity->isPlayer()) {
            entity->remove();
            return "1\n";
        }
        return "0\n";
    } else if (cmd == "world.removeEntities") {
        int type;
        int ret = sscanf(args.c_str(), "%d", &type);
        if (ret != 1) return "0\n";

        int removed = 0;
        for (Entity *entity : commandserver->minecraft->level->entities) {
            if ((type == -1 || entity->getEntityTypeId() == type) && !entity->isPlayer()) {
                entity->remove();
                removed++;
            }
        }
        return std::to_string(removed) + "\n";
    } else if (cmd == "events.chat.posts") {
        std::string ret = "";
        while (chatEvents.at != chatEvents.start) {
            ChatEvent ce = chatEvents.pop();
            if (!ce.invalid) {
                std::string message = ce.toString(commandserver);
                if (compat_mode) std::replace(message.begin(), message.end(), '|', '\\');
                else message = misc_base64_encode(message);
                ret += message + "|";
            }
        }
        if (ret.size() > 1) ret.pop_back();
        return ret + "\n";
    } else if (cmd == "player.events.chat.posts") {
        if (commandserver->minecraft->player == NULL) return fail;
        return getEventsOfId(chatEvents, commandserver, commandserver->minecraft->player->id);
    } else if (cmd == "entity.events.chat.posts") {
        int id;
        int ret = sscanf(args.c_str(), "%d", &id);
        if (ret != 1) return fail;
        return getEventsOfId(chatEvents, commandserver, id);
    } else if (cmd == "player.events.block.hits") {
        if (commandserver->minecraft->player == NULL) return fail;
        return getBlockHits(commandserver, client, commandserver->minecraft->player->id);
    } else if (cmd == "entity.events.block.hits") {
        int id;
        int ret = sscanf(args.c_str(), "%d", &id);
        if (ret != 1) return fail;
        return getBlockHits(commandserver, client, id);
    } else if (cmd == "events.projectile.hits") {
        std::string result = "";
        while (hitEvents.at != hitEvents.start) {
            ProjectileHitEvent he = hitEvents.pop();
            if (!he.invalid)
                result += he.toString(commandserver) + "|";
        }
        if (result.size() > 1) result.pop_back();
        return result + "\n";
    } else if (cmd == "player.events.projectile.hits") {
        if (commandserver->minecraft->player == NULL) return fail;
        return getEventsOfId(hitEvents, commandserver, commandserver->minecraft->player->id);
    } else if (cmd == "entity.events.projectile.hits") {
        int id;
        int ret = sscanf(args.c_str(), "%d", &id);
        if (ret != 1) return fail;
        return getEventsOfId(hitEvents, commandserver, id);
    } else if (cmd == "player.setDirection" && commandserver->minecraft->player) {
        if (commandserver->minecraft->player == NULL) return fail;
        float x, y, z;
        int ret = sscanf(args.c_str(), "%f,%f,%f", &x, &y, &z);
        if (ret != 3) return "";
        setDir((Entity *) commandserver->minecraft->player, x, y, z);
    } else if (cmd == "entity.setDirection") {
        int id;
        float x, y, z;
        int ret = sscanf(args.c_str(), "%d,%f,%f,%f", &id, &x, &y, &z);
        if (ret != 4) return "";
        Entity *entity = commandserver->minecraft->level->getEntity(id);
        if (entity == NULL) {
            WARN("Entity [%i] not found.", id);
        } else {
            setDir(entity, x, y, z);
        }
    } else if (cmd == "player.getDirection" && commandserver->minecraft->player) {
        Vec3 vec = getDir((Entity *) commandserver->minecraft->player);
        return std::to_string(vec.x) + "," + std::to_string(vec.y) + "," + std::to_string(vec.z) + "\n";
    } else if (cmd == "entity.getDirection") {
        int id;
        int ret = sscanf(args.c_str(), "%d", &id);
        if (ret != 1) return "";
        Entity *entity = commandserver->minecraft->level->getEntity(id);
        if (entity == NULL) {
            WARN("Entity [%i] not found.", id);
            return fail;
        } else {
            Vec3 vec = getDir(entity);
            return std::to_string(vec.x) + "," + std::to_string(vec.y) + "," + std::to_string(vec.z) + "\n";
        }
    } else if (cmd == "player.setRotation"  && commandserver->minecraft->player) {
        float yaw;
        int ret = sscanf(args.c_str(), "%f", &yaw);
        if (ret != 1) return "";
        commandserver->minecraft->player->yaw = yaw;
    } else if (cmd == "entity.setRotation") {
        int id;
        float yaw;
        int ret = sscanf(args.c_str(), "%d,%f", &id, &yaw);
        if (ret != 2) return "";
        Entity *entity = commandserver->minecraft->level->getEntity(id);
        if (entity == NULL) {
            WARN("Entity [%i] not found.", id);
        } else {
            entity->yaw = yaw;
        }
    } else if (cmd == "player.setPitch" && commandserver->minecraft->player) {
        float pitch;
        int ret = sscanf(args.c_str(), "%f", &pitch);
        if (ret != 1) return "";
        commandserver->minecraft->player->pitch = pitch;
    } else if (cmd == "entity.setPitch") {
        int id;
        float pitch;
        int ret = sscanf(args.c_str(), "%d,%f", &id, &pitch);
        if (ret != 2) return "";
        Entity *entity = commandserver->minecraft->level->getEntity(id);
        if (entity == NULL) {
            WARN("Entity [%i] not found.", id);
        } else {
            entity->pitch = pitch;
        }
    } else if (cmd == "player.getRotation" && commandserver->minecraft->player) {
        return std::to_string(commandserver->minecraft->player->yaw) + "\n";
    } else if (cmd == "entity.getRotation") {
        int id;
        int ret = sscanf(args.c_str(), "%d", &id);
        if (ret != 1) return "";
        Entity *entity = commandserver->minecraft->level->getEntity(id);
        if (entity == NULL) {
            WARN("Entity [%i] not found.", id);
            return fail;
        } else {
            return std::to_string(entity->yaw) + "\n";
        }
    } else if (cmd == "player.getPitch" && commandserver->minecraft->player) {
        return std::to_string(commandserver->minecraft->player->pitch) + "\n";
    } else if (cmd == "entity.getPitch") {
        int id;
        int ret = sscanf(args.c_str(), "%d", &id);
        if (ret != 1) return "";
        Entity *entity = commandserver->minecraft->level->getEntity(id);
        if (entity == NULL) {
            WARN("Entity [%i] not found.", id);
            return fail;
        } else {
            return std::to_string(entity->pitch) + "\n";
        }
    } else if (cmd == "player.getEntities" || cmd == "entity.getEntities") {
        // Parse
        int dist, type;
        Entity *src = NULL;
        if (cmd == "player.getEntities") {
            int ret = sscanf(args.c_str(), "%d,%d", &dist, &type);
            if (ret != 2 || commandserver->minecraft->player == NULL) return "";
            src = (Entity *) commandserver->minecraft->player;
        } else {
            int id = 0;
            int ret = sscanf(args.c_str(), "%d,%d,%d", &id, &dist, &type);
            if (ret != 3) return "";
            src = commandserver->minecraft->level->getEntity(id);
            if (src == NULL) {
                WARN("Entity [%i] not found.", id);
                return fail;
            }
        }
        // Run
        std::string result = "";
        for (Entity *entity : commandserver->minecraft->level->entities) {
            if ((type == -1 || entity->getEntityTypeId() == type) && distance_between(src, entity) < dist) {
                result += getEntityData(commandserver, entity) + "|";
            }
        }
        if (result.size() > 1) result.pop_back();
        return result + "\n";
    } else if (cmd == "player.removeEntities" || cmd == "entity.removeEntities") {
        // Parse
        int dist, type;
        Entity *src = NULL;
        if (cmd == "player.removeEntities") {
            int ret = sscanf(args.c_str(), "%d,%d", &dist, &type);
            if (ret != 2 || commandserver->minecraft->player == NULL) return "";
            src = (Entity *) commandserver->minecraft->player;
        } else {
            int id = 0;
            int ret = sscanf(args.c_str(), "%d,%d,%d", &id, &dist, &type);
            if (ret != 3) return "";
            src = commandserver->minecraft->level->getEntity(id);
            if (src == NULL) {
                WARN("Entity [%i] not found.", id);
                return fail;
            }
        }
        // Run
        int removed = 0;
        for (Entity *entity : commandserver->minecraft->level->entities) {
            if ((type == -1 || entity->getEntityTypeId() == type) && distance_between(src, entity) < dist && !entity->isPlayer()) {
                entity->remove();
                removed++;
            }
        }
        return std::to_string(removed) + "\n";
    } else if (cmd == "world.setSign") {
        // Parse
        int x, y, z, id, data;
        char l1[100], l2[100], l3[100], l4[100];
        int ret = sscanf(args.c_str(), "%d,%d,%d,%d,%d,%99[^,],%99[^,],%99[^,],%99s", &x, &y, &z, &id, &data, l1, l2, l3, l4);
        if (ret < 5) return fail;
        // Translate
        commandserver->pos_translator.from(x, y, z);
        // Set block
        commandserver->minecraft->level->setTileAndData(x, y, z, id, data);
        // Set sign data
        if (ret == 3) return "";
        SignTileEntity *sign = (SignTileEntity *) commandserver->minecraft->level->getTileEntity(x, y, z);
        if (sign == NULL || sign->type != 4) return "";
        if (ret > 5) sign->lines[0] = compat_mode ? l1 : misc_base64_decode(l1);
        if (ret > 6) sign->lines[1] = compat_mode ? l2 : misc_base64_decode(l2);
        if (ret > 7) sign->lines[2] = compat_mode ? l3 : misc_base64_decode(l3);
        if (ret > 8) sign->lines[3] = compat_mode ? l4 : misc_base64_decode(l4);
    } else if (cmd == "world.spawnEntity") {
        // Parse
        float x, y, z;
        int id;
        int ret = sscanf(args.c_str(), "%f,%f,%f,%d", &x, &y, &z, &id);
        if (ret != 4) return fail;
        // Translate
        int ix = x, iy = y, iz = z;
        int ix1 = x, iy1 = y, iz1 = z;
        commandserver->pos_translator.from(ix, iy, iz);
        x += ix - ix1;
        y += iy - iy1;
        z += iz - iz1;
        if (compat_mode && modern_entity_id_mapping.contains(id)) id = modern_entity_id_mapping[id];
        // Spawn
        Entity *entity = misc_make_entity_from_id(commandserver->minecraft->level, id);
        if (entity == NULL) return fail;
        entity->moveTo(x, y, z, 0, 0);
        commandserver->minecraft->level->addEntity(entity);
        return std::to_string(entity->id) + "\n";
    } else if (cmd == "world.getEntityTypes") {
        std::string result = "";
        for (auto &i : misc_get_entity_names()) {
            if (compat_mode && modern_entity_id_mapping.contains(i.first)) result += std::to_string(modern_entity_id_mapping[i.first]);
            else result += std::to_string(i.first);
            result += "," + i.second + "|";
        }
        if (result.size() > 1) result.pop_back();
        for (auto &c : result) c = toupper(c);
        return result + "\n";
    } else if (cmd == "player.setAbsPos" && commandserver->minecraft->player) {
        float x, y, z;
        int ret = sscanf(args.c_str(), "%f,%f,%f", &x, &y, &z);
        if (ret != 3) return fail;
        commandserver->minecraft->player->moveTo(x, y, z, commandserver->minecraft->player->yaw, commandserver->minecraft->player->pitch);
    } else if (cmd == "player.getAbsPos" && commandserver->minecraft->player) {
        return std::to_string(commandserver->minecraft->player->x) + ","
            + std::to_string(commandserver->minecraft->player->y) + ","
            + std::to_string(commandserver->minecraft->player->z) + "\n";
    } else if (cmd == "player.events.clear") {
        if (commandserver->minecraft->player == NULL) return "";
        clearEvents(commandserver, commandserver->minecraft->player->id);
    } else if (cmd == "entity.events.clear") {
        int id;
        int ret = sscanf(args.c_str(), "%d", &id);
        if (ret != 1) return fail;
        clearEvents(commandserver, id);
    } else if (cmd == "events.clear") {
        clearNewEvents();
        // Clear the other events as well
        return old(commandserver, client, command);
    } else if (cmd == "reborn.disableCompatMode") {
        compat_mode = false;
    } else if (cmd == "reborn.enableCompatMode") {
        compat_mode = true;
    } else {
        // Either invalid or a vanilla command, either way hand it off to the orignal handler
        return old(commandserver, client, command);
    }
    // No return, it's different from empty! Empty is "\n". "" will result in nothing being sent back while "\n" will result in an empty string being sent back
    return "";
}

// Arrow entity hit
static Entity *shooter = NULL;
static HitResult *Arrow_tick_HitResult_constructor_injection(HitResult *self, Entity *target) {
    // Orignal
    self->constructor(target);
    // Add event
    if (shooter && shooter->isPlayer()) {
        hitEvents.push(ProjectileHitEvent{
            .x = (int) target->x,
            .y = (int) target->y,
            .z = (int) target->z,
            .owner = misc_get_player_username((Player *) shooter),
            .owner_id = shooter->id,
            .targetId = target->id
        });
    }
    return self;
}

// Arrow block hit
static void Arrow_tick_injection(Arrow_tick_t old, Arrow *self) {
    int oldFlightTime = self->flight_time;
    shooter = self->level->getEntity(self->getAuxData());
    old(self);
    if (self && !self->pending_removal && self->grounded && oldFlightTime != self->flight_time) {
        if (shooter && shooter->isPlayer()) {
            // Hit! Get the data
            hitEvents.push(ProjectileHitEvent{
                .x = self->hit_x,
                .y = self->hit_y,
                .z = self->hit_z,
                .owner = misc_get_player_username((Player *) shooter),
                .owner_id = shooter->id,
                .targetId = 0
            });
        }
    }
}

// Throwable hits
static void Throwable_tick_Throwable_onHit_injection(Throwable *self, HitResult *res) {
    if (res == NULL || res->type == 2) return self->onHit(res);
    Entity *thrower = self->level->getEntity(self->getAuxData());
    if (thrower == NULL || !thrower->isPlayer()) return self->onHit(res);
    ProjectileHitEvent event;
    if (res->type == 1 && res->entity) {
        // Entity
        event = ProjectileHitEvent {
            .x = (int) res->exact.x,
            .y = (int) res->exact.y,
            .z = (int) res->exact.z,
            .owner = misc_get_player_username((Player *) thrower),
            .owner_id = thrower->id,
            .targetId = res->entity->id
        };
    } else {
        // Tile
        event = ProjectileHitEvent {
            .x = res->x,
            .y = res->y,
            .z = res->z,
            .owner = misc_get_player_username((Player *) thrower),
            .owner_id = thrower->id,
            .targetId = 0
        };
    }
    hitEvents.push(event);
    self->onHit(res);
}

static void Gui_addMessage_injection(Gui_addMessage_t original, Gui *gui, const std::string &text) {
    static bool recursing = false;
    if (recursing || chat_already_added()) {
        chat_set_already_added(false);
        original(gui, text);
    } else {
        if (chat_is_sending()) {
            std::string message = text.substr(gui->minecraft->player->username.size() + 3);
            api_add_chat_event((Entity *) gui->minecraft->player, text);
        } else {
            api_add_chat_event(NULL, text);
        }
        recursing = true;
        original(gui, text);
        recursing = false;
    }
}

bool enabled = false;
void api_add_chat_event(Entity *sender, const std::string &message) {
    if (!enabled) return;
    chatEvents.push(ChatEvent{message, sender ? sender->id : 0});
}

void init_api() {
    if (feature_has("Implement RaspberryJuice API", server_enabled)) {
        enabled = true;
        overwrite_calls(CommandServer_parse, CommandServer_parse_injection);
        overwrite_calls(Arrow_tick, Arrow_tick_injection);
        overwrite_call((void *) 0x8b1e8, (void *) Arrow_tick_HitResult_constructor_injection);
        overwrite_call((void *) 0x8c5a4, (void *) Throwable_tick_Throwable_onHit_injection);
        overwrite_calls(Gui_addMessage, Gui_addMessage_injection);
    }
}
